"""Auxiliary functions for the pounders algorithm."""
from typing import NamedTuple
from typing import Union

import numpy as np
from estimagic.optimization.subsolvers.quadratic_subsolvers import (
    minimize_bntr_quadratic,
)
from estimagic.optimization.subsolvers.quadratic_subsolvers import (
    minimize_gqtpar_quadratic,
)
from scipy.linalg import qr_multiply


class ResidualModel(NamedTuple):
    intercepts: Union[np.ndarray, None] = None  # shape (n_residuals,)
    linear_terms: Union[np.ndarray, None] = None  # shape (n_residuals, n_params)
    square_terms: Union[
        np.ndarray, None
    ] = None  # shape (n_residuals, n_params, n_params)


class MainModel(NamedTuple):
    linear_terms: Union[np.ndarray, None] = None  # shape (n_params,)
    square_terms: Union[np.ndarray, None] = None  # shape (n_params, n_params)


def create_initial_residual_model(history, accepted_index, delta):
    """Update linear and square terms of the initial residual model.

    Args:
        history (class): Class storing history of xs, residuals, and critvals.
        accepted_index (int): Index in history pointing to the currently
            accepted candidate vector.
        delta (float): Trust-region radius.

    Returns:
        NamedTuple: NamedTuple containing the parameters of the residual model
            with updated ``linear_terms`` and ``square_terms``.
    """
    center_info = {
        "x": history.get_best_x(),
        "residuals": history.get_best_residuals(),
        "radius": delta,
    }
    n = len(center_info["x"])
    n_obs = center_info["residuals"].shape[0]

    indices_not_min = [i for i in range(n + 1) if i != accepted_index]

    x_candidate, residuals_candidate, _ = history.get_centered_entries(
        center_info=center_info,
        index=indices_not_min,
    )

    linear_terms = np.linalg.solve(x_candidate, residuals_candidate)
    square_terms = np.zeros((n_obs, n, n))

    residual_model = ResidualModel(
        intercepts=history.get_best_residuals(),
        linear_terms=linear_terms,
        square_terms=square_terms,
    )

    return residual_model


def update_residual_model(residual_model, coefficients_to_add, delta, delta_old):
    """Update linear and square terms of the residual model.

    Args:
        residual_model (NamedTuple): NamedTuple containing the parameters of
            residual model, i.e. ``intercepts``, ``linear_terms``, and ``square terms``.
        coefficients_to_add (dict): Coefficients used for updating the
            parameters of the residual model.
        delta (float): Trust region radius of the current iteration.
        delta_old (float): Trust region radius of the previous iteration.

    Returns:
        NamedTuple: NamedTuple containing the parameters of the residual model
            with update ``linear_terms`` and ``square_terms``.
    """
    linear_terms_new = (
        coefficients_to_add["linear_terms"]
        + (delta / delta_old) * residual_model.linear_terms
    )

    square_terms_new = (
        coefficients_to_add["square_terms"]
        + (delta / delta_old) ** 2 * residual_model.square_terms
    )

    residual_model_updated = residual_model._replace(
        linear_terms=linear_terms_new, square_terms=square_terms_new
    )

    return residual_model_updated


def create_main_from_residual_model(
    residual_model, multiply_square_terms_with_residuals=True
):
    """Update linear and square terms of the main model via the residual model.

    Args:
        residual_model (NamedTuple): NamedTuple containing the parameters of
            residual model, i.e. "intercepts", ``linear_terms``, and "square terms".
        multiply_square_terms_with_residuals (bool): Indicator whether we
            multiply the main model's "square terms" with residuals, i.e.
            the intercepts of the residual model.

    Returns:
        NamedTuple: NamedTuple containing the updated parameters of the
            main model, i.e. ``linear_terms`` and ``square terms``.
    """
    linear_terms_main_model = np.dot(
        residual_model.linear_terms, residual_model.intercepts
    )
    square_terms_main_model = np.dot(
        residual_model.linear_terms, residual_model.linear_terms.T
    )

    if multiply_square_terms_with_residuals is True:
        # Multiply 3d array *square_terms_residual_model* with
        # 1d array *intercepts_residual_model* along axis 0 of the former.
        dim_array = np.ones((1, residual_model.square_terms.ndim), int).ravel()
        dim_array[0] = -1

        intercepts_reshaped = residual_model.intercepts.reshape(dim_array)

        square_terms_main_model = square_terms_main_model + np.sum(
            intercepts_reshaped * residual_model.square_terms, axis=0
        )

    main_model = MainModel(
        linear_terms=linear_terms_main_model, square_terms=square_terms_main_model
    )

    return main_model


def update_main_model_with_new_accepted_x(main_model, x_candidate):
    """Use accepted candidate to update the linear terms of the residual model.

    Args:
         main_model (NamedTuple): NamedTuple containing the parameters of the
            main model, i.e. ``linear_terms`` and ``square terms``.
        x_candidate (np.ndarray): Vector of centered x candidates of shape (n,).

    Returns:
        NamedTuple: NamedTuple containing the parameters of the main model
            with updated ``linear_terms``.
    """
    linear_terms_new = main_model.linear_terms + np.dot(
        main_model.square_terms, x_candidate
    )
    main_model_updated = main_model._replace(linear_terms=linear_terms_new)

    return main_model_updated


def update_residual_model_with_new_accepted_x(residual_model, x_candidate):
    """Use accepted candidate to update residual model.

    Args:
        residual_model (NamedTuple): NamedTuple containing the parameters of
            the residual model, i.e. ``intercepts``, ``linear_terms``, and
            ``square terms``.
        x_candidate (np.ndarray): Vector of centered x candidates of shape (n,).

    Returns:
        NamedTuple: NamedTuple containing the parameters of the residual model
            with updated ``intercepts`` and ``linear_terms``.
    """
    intercepts_new = (
        residual_model.intercepts
        + np.dot(x_candidate, residual_model.linear_terms)
        + 0.5 * np.dot(np.dot(x_candidate, residual_model.square_terms), x_candidate)
    )
    linear_terms_new = (
        residual_model.linear_terms + np.dot(residual_model.square_terms, x_candidate).T
    )

    residual_model_updated = residual_model._replace(
        intercepts=intercepts_new, linear_terms=linear_terms_new
    )
    return residual_model_updated


def solve_subproblem(
    x_accepted,
    main_model,
    lower_bounds,
    upper_bounds,
    delta,
    solver,
    *,
    conjugate_gradient_method,
    maxiter,
    maxiter_gradient_descent,
    gtol_abs,
    gtol_rel,
    gtol_scaled,
    gtol_abs_conjugate_gradient,
    gtol_rel_conjugate_gradient,
    k_easy,
    k_hard
):
    """Solve the quadratic subproblem.

    Args:
        x_accepted (np.ndarray): Currently accepted candidate vector of shape (n,).
        delta (float): Current trust region radius.
        main_model (NamedTuple): NamedTuple containing the parameters of the
            main model, i.e. ``linear_terms`` and ``square terms``.
        lower_bounds (np.ndarray): 1d array of shape (n,) with lower bounds
            for the parameter vector x.
        upper_bounds (np.ndarray): 1d array of shape (n,) with upper bounds
            for the parameter vector x.
        delta (float) Current trust-region radius
        solver (str): Trust-region subsolver to use. Currently, two internal solvers
            are supported:
            - "bntr" (default, supports bound constraints)
            - "gqtpar" (does not support bound constraints)
        conjugate_gradient_method (str): Method for computing the conjugate gradient
            step. Available conjugate gradient methods are:
                - "cg"
                - "steihaug_toint"
                - "trsbox" (default)
        maxiter (int): Maximum number of iterations to perform when solving the
            trust-region subproblem.
        maxiter_gradient_descent (int): Maximum number of gradient descent iterations
            to perform when the trust-region subsolver "bntr" is used.
        gtol_abs (float): Convergence tolerance for the absolute gradient norm
            in the trust-region subproblem ("bntr").
        gtol_rel (float): Convergence tolerance for the relative gradient norm
            in the trust-region subproblem ("bntr").
        gtol_scaled (float): Convergence tolerance for the scaled gradient norm
            in the trust-region subproblem ("bntr").
        gtol_abs_conjugate_gradient (float): Convergence tolerance for the absolute
            gradient norm in the conjugate gradient step of the trust-region
            subproblem ("bntr").
        gtol_rel_conjugate_gradient (float): Convergence tolerance for the relative
            gradient norm in the conjugate gradient step of the trust-region
            subproblem ("bntr").
        k_easy (float): topping criterion for the "easy" case in the trust-region
            subproblem ("gqtpar").
        k_hard (float): Stopping criterion for the "hard" case in the trust-region
            subproblem ("gqtpar").

    Returns:
        (dict): Result dictionary containing the followng keys:
            - "x" (np.ndarray): The solution vector of shape (n,)
            - "criterion" (float): The value of the criterion functions associated
                with the solution
            - "n_iterations" (int): Number of iterations performed before termination.
            - "success" (bool): Boolean indicating whether a solution has been found
                before reaching maxiter.
    """
    # Initial guess
    x0 = np.zeros_like(x_accepted)

    # Normalize bounds. If none provided, use unit cube [-1, 1]
    if lower_bounds is not None:
        lower_bounds = (lower_bounds - x_accepted) / delta
        lower_bounds[lower_bounds < -1] = -1
    else:
        lower_bounds = -np.ones_like(x_accepted)

    if upper_bounds is not None:
        upper_bounds = (upper_bounds - x_accepted) / delta
        upper_bounds[upper_bounds > 1] = 1
    else:
        upper_bounds = np.ones_like(x_accepted)

    # Check if bounds valid
    if np.max(lower_bounds - upper_bounds) > 1e-10:
        raise ValueError("Upper bounds < lower bounds in subproblem.")
    if np.max(lower_bounds - x0) > 1e-10:
        raise ValueError("Initial guess < lower bounds in subproblem.")
    if np.max(x0 - upper_bounds) > 1e-10:
        raise ValueError("Initial guess > upper bounds in subproblem.")

    if solver == "bntr":
        options = {
            "conjugate_gradient_method": conjugate_gradient_method,
            "maxiter": maxiter,
            "maxiter_gradient_descent": maxiter_gradient_descent,
            "gtol_abs": gtol_abs,
            "gtol_rel": gtol_rel,
            "gtol_scaled": gtol_scaled,
            "gtol_abs_conjugate_gradient": gtol_abs_conjugate_gradient,
            "gtol_rel_conjugate_gradient": gtol_rel_conjugate_gradient,
        }
        result = minimize_bntr_quadratic(
            main_model, lower_bounds, upper_bounds, **options
        )
    elif solver == "gqtpar":
        result = minimize_gqtpar_quadratic(
            main_model,
            k_easy=k_easy,
            k_hard=k_hard,
            maxiter=maxiter,
        )
    else:
        raise ValueError(
            "Invalid subproblem solver: {solver}. Must be one of bntr, gqtpar."
        )

    # Test bounds post-solution
    if np.max(lower_bounds - result["x"]) > 1e-5:
        raise ValueError("Subproblem solution < lower bounds.")
    if np.max(result["x"] - upper_bounds) > 1e-5:
        raise ValueError("Subproblem solution > upper bounds.")

    return result


def find_affine_points(
    history,
    x_accepted,
    model_improving_points,
    project_x_onto_null,
    delta,
    theta1,
    c,
    model_indices,
    n_modelpoints,
):
    """Find affine points.

    Args:
        history (class): Class storing history of xs, residuals, and critvals.
        x_accepted (np.ndarray): Accepted solution vector of the subproblem.
            Shape (n,).
        model_improving_points (np.ndarray): Array of shape (n, n) including
            points to improve the main model, i.e. make the main model fully linear.
            If *project_x_onto_null* is False, it is an array filled with zeros.
        project_x_onto_null (int): Indicator whether to calculate the QR
            decomposition of *model_improving_points* and multiply it
             with vector *x_projected*.
        delta (float): Delta, current trust-region radius.
        theta1 (float): Threshold for adding the current x candidate to the model.
        c (float): Threshold for acceptance of the norm of our current x candidate.
        model_indices (np.ndarray): Indices related to the candidates of x
            that are currently in the main model. Shape (2 *n* + 1,).
        n_modelpoints (int): Current number of model points.

    Returns:
        Tuple:
        - model_improving_points (np.ndarray):  Array of shape (n, n) including
            points to improve the main model, i.e. make the main model fully linear.
        - model_indices (np.ndarray): Indices related to the candidates of x
            that are currently in the main model. Shape (2 *n* + 1,).
        - n_modelpoints (int): Current number of model points.
        - project_x_onto_null (int): Indicator whether to calculate the QR
            decomposition of *model_improving_points* and multiply it
            with vector *x_projected*.
            Relevant for next call of *find_affine_points()*.
    """
    n = len(x_accepted)

    for i in range(history.get_n_fun() - 1, -1, -1):
        center_info = {"x": x_accepted, "radius": delta}
        x_candidate = history.get_centered_xs(center_info, index=i)
        candidate_norm = np.linalg.norm(x_candidate)

        x_projected = x_candidate

        if candidate_norm <= c:
            if project_x_onto_null is True:
                x_projected, _ = qr_multiply(model_improving_points, x_projected)

            proj = np.linalg.norm(x_projected[n_modelpoints:])

            # Add this index to the model
            if proj >= theta1:
                model_indices[n_modelpoints] = i
                model_improving_points[:, n_modelpoints] = x_candidate
                project_x_onto_null = True
                n_modelpoints += 1

            if n_modelpoints == n:
                break

    return model_improving_points, model_indices, n_modelpoints, project_x_onto_null


def add_geomtery_points_to_make_main_model_fully_linear(
    history,
    main_model,
    model_improving_points,
    model_indices,
    x_accepted,
    n_modelpoints,
    delta,
    criterion,
    lower_bounds,
    upper_bounds,
    batch_evaluator,
    n_cores,
):
    """Add points until main model is fully linear.

    Args:
        history (class): Class storing history of xs, residuals, and critvals.
        main_model (NamedTuple): NamedTuple containing the parameters of the
            main model, i.e. ``linear_terms`` and "square terms".
        model_improving_points (np.ndarray): Array of shape (n, n) including
            points to improve the main model.
        model_indices (np.ndarray): Indices of the candidates of x that are
            currently in the main model. Shape (2 * n + 1,).
        x_accepted (np.ndarray): Accepted solution vector of the subproblem.
            Shape (n,).
        n_modelpoints (int): Current number of model points.
        delta (float): Delta, current trust-region radius.
        criterion (callable): Criterion function.
        lower_bounds (np.ndarray): Lower bounds.
            Must have same length as the initial guess of the
            parameter vector. Equal to -1 if not provided by the user.
        upper_bounds (np.ndarray): Upper bounds.
            Must have same length as the initial guess of the
            parameter vector. Equal to 1 if not provided by the user.
        batch_evaluator (str or callable): Name of a pre-implemented batch evaluator
            (currently 'joblib' and 'pathos_mp') or Callable with the same interface
            as the estimagic batch_evaluators.
        n_cores (int): Number of processes used to parallelize the function
            evaluations.

    Returns:
        Tuple:
        - history (class): Class storing history of xs, residuals, and critvals.
        - model_indices (np.ndarray): Indices of the candidates of x that are
            currently in the main model. Shape (2 * n + 1,).
    """
    n = len(x_accepted)

    current_history = history.get_n_fun()

    x_candidate = np.zeros_like(x_accepted)
    x_candidates_list = []
    criterion_candidates_list = []

    model_improving_points, _ = qr_multiply(model_improving_points, np.eye(n))

    for i in range(n_modelpoints, n):
        change_direction = np.dot(model_improving_points[:, i], main_model.linear_terms)

        if change_direction > 0:
            model_improving_points[:, i] *= -1

        x_candidate = delta * model_improving_points[:, i] + x_accepted

        # Project into feasible region
        if lower_bounds is not None and upper_bounds is not None:
            x_candidate = np.median(
                np.stack([lower_bounds, x_candidate, upper_bounds]), axis=0
            )
        x_candidates_list.append(x_candidate)
        model_indices[i] = current_history + i - n_modelpoints

    criterion_candidates_list = batch_evaluator(
        criterion, arguments=x_candidates_list, n_cores=n_cores
    )

    history.add_entries(x_candidates_list, criterion_candidates_list)

    return history, model_indices


def get_interpolation_matrices_residual_model(
    history,
    x_accepted,
    model_indices,
    delta,
    c2,
    theta2,
    n_maxinterp,
    n_modelpoints,
):
    """Obtain matrices that will be used for interpolating the residual model.

    Args:
        history (class): Class storing history of xs, residuals, and critvals.
        x_accepted (np.ndarray): Accepted solution vector of the subproblem.
            Shape (n,).
        model_indices (np.ndarray): Indices of the candidates of x that are
            currently in the model. Shape (2 * n + 1,).
        delta (float): Delta, current trust-region radius.
        c2 (int): Threshold for acceptance of the norm of our current x candidate.
            Equal to 10 by default.
        theta2 (float): Threshold for adding the current x candidate to the model.
        n_maxinterp (int): Maximum number of interpolation points.
        n_modelpoints (int): Current number of model points.

    Returns:
        Tuple:
        - x_sample_monomial_basis (np.ndarray): Sample of xs used for
            building the monomial basis. When taken together, they
            form a basis for the linear space of quadratics in n
            variables.
            Shape(n_maxinterp, n * (n + 1) / 2).
        - monomial_basis (np.ndarray): Monomial basis for quadratic functions of x.
            Shape(n_maxinterp, n * (n + 1) / 2).
        - basis_null_space (np.ndarray): Basis for the null space of xs that
            form the monomial basis. Shape(n_maxinterp, len(n + 1 : n_modelpoints)).
        - lower_triangular (np.ndarray): Lower triangular matrix of xs that
            form the monomial basis. Shape(n_maxinterp, n * (n + 1) / 2).
        - n_modelpoints (int): Current number of model points.
    """
    n = len(x_accepted)

    x_sample_monomial_basis = np.zeros((n_maxinterp, n + 1))
    x_sample_monomial_basis[:, 0] = 1
    x_sample_full_with_zeros = np.zeros((n_maxinterp, n_maxinterp))
    x_sample_full_with_zeros[:n_maxinterp, : n + 1] = x_sample_monomial_basis

    monomial_basis = np.zeros((n_maxinterp, int(n * (n + 1) / 2)))

    center_info = {"x": x_accepted, "radius": delta}
    for i in range(n + 1):
        x_sample_monomial_basis[i, 1:] = history.get_centered_xs(
            center_info, index=model_indices[i]
        )
        monomial_basis[i, :] = _get_monomial_basis(x_sample_monomial_basis[i, 1:])

    # Now we add points until we have n_maxinterp starting with the most recent ones
    point = history.get_n_fun() - 1
    n_modelpoints = n + 1

    while (n_modelpoints < n_maxinterp) and (point >= 0):
        reject = False

        # Reject any points already in the model
        for i in range(n + 1):
            if point == model_indices[i]:
                reject = True
                break

        if reject is False:
            candidate_x = history.get_centered_xs(center_info, index=point)
            candidate_norm = np.linalg.norm(candidate_x)

            if candidate_norm > c2:
                reject = True

        if reject is True:
            point -= 1
            continue

        x_sample_monomial_basis[n_modelpoints, 1:] = history.get_centered_xs(
            center_info, index=point
        )
        monomial_basis[n_modelpoints, :] = _get_monomial_basis(
            x_sample_monomial_basis[n_modelpoints, 1:]
        )

        x_sample_full_with_zeros = np.zeros((n_maxinterp, n_maxinterp))
        x_sample_full_with_zeros[:n_maxinterp, : n + 1] = x_sample_monomial_basis

        lower_triangular_temporary, _ = qr_multiply(
            x_sample_full_with_zeros[: n_modelpoints + 1, :],
            monomial_basis.T[: int(n * (n + 1) / 2), : n_modelpoints + 1],
        )
        beta = np.linalg.svd(lower_triangular_temporary.T[n + 1 :], compute_uv=False)

        if beta[min(n_modelpoints - n, int(n * (n + 1) / 2)) - 1] > theta2:
            # Accept point
            model_indices[n_modelpoints] = point
            lower_triangular = lower_triangular_temporary

            n_modelpoints += 1

        point -= 1

    # Orthogonal basis for the null space of M, where M is the
    # sample of xs forming the monomial basis
    basis_null_space, _ = qr_multiply(
        x_sample_full_with_zeros[:n_modelpoints, :],
        np.eye(n_maxinterp)[:, :n_modelpoints],
    )
    basis_null_space = basis_null_space[:, n + 1 : n_modelpoints]

    if n_modelpoints == (n + 1):
        lower_triangular = np.zeros((n_maxinterp, int(n * (n + 1) / 2)))
        lower_triangular[:n, :n] = np.eye(n)

    return (
        x_sample_monomial_basis,
        monomial_basis,
        basis_null_space,
        lower_triangular,
        n_modelpoints,
    )


def interpolate_residual_model(
    history,
    interpolation_set,
    residual_model,
    model_indices,
    n_modelpoints,
    n_maxinterp,
):
    """Interpolate the quadratic residual model.

    The residual model:

        Q(x) = c + g'x + 0.5 x G x'

    satisfies the interpolation conditions Q(X[:,j]) = f(j)
    for j= 1,..., m with a Hessian matrix of least Frobenius norm.

    If the point x_k belongs to the interpolation set, one can show that
    c = f (x_k).

    Args:
        history (class): Class storing history of xs, residuals, and critvals.
        x_sample (np.ndarray): Vector of centered x sample that makes up the
            interpolation set. Shape (maxinterp, n).
        residual_model (NamedTuple): NamedTuple containing the parameters of
            residual model, i.e. ``intercepts``, ``linear_terms``, and ``square terms``.
        model_indices (np.ndarray): Indices of the candidates of x that are
            currently in the model. Shape (2 *n* + 1,).
        n_modelpoints (int): Current number of model points.

    Returns:
        np.ndarray: Interpolated residual model. Array of shape
            (n_maxinterp, n_obs).
    """
    n_obs = history.get_residuals(index=-1).shape[0]
    residual_model_interpolated = np.zeros((n_maxinterp, n_obs), dtype=np.float64)

    for j in range(n_obs):
        x_dot_square_terms = np.dot(
            interpolation_set, residual_model.square_terms[j, :, :]
        )

        for i in range(n_modelpoints):
            center_info = {"residuals": residual_model.intercepts}
            residuals = history.get_centered_residuals(
                center_info, index=model_indices[i]
            )

            residual_model_interpolated[i, j] = (
                residuals[j]
                - np.dot(residual_model.linear_terms[:, j], interpolation_set[i, :])
                - 0.5 * np.dot(x_dot_square_terms[i, :], interpolation_set[i, :])
            )

    return residual_model_interpolated


def get_coefficients_residual_model(
    lower_triangular,
    basis_null_space,
    monomial_basis,
    x_sample_monomial_basis,
    residual_model_interpolated,
    n_modelpoints,
):
    """Computes the coefficients of the quadratic residual model.

    Args:
        x_sample_monomial_basis (np.ndarray): Sample of xs used for
            building the monomial basis. When taken together, they
            form a basis for the linear space of quadratics in n
            variables.
            Shape(n_maxinterp, n * (n + 1) / 2).
        monomial_basis (np.ndarray): Monomial basis for quadratic functions of x.
            Shape(n_maxinterp, n * (n + 1) / 2).
        basis_null_space (np.ndarray): Basis for the null space of xs that
            form the monomial basis. Shape(n_maxinterp, len(n + 1 : n_modelpoints)).
        lower_triangular (np.ndarray): Lower triangular matrix of xs that
            form the monomial basis. Shape(n_maxinterp, n * (n + 1) / 2).
        f_interpolated (np.ndarray): Interpolated criterion function f.
            Shape (n_maxinterp, n_obs).
        n_modelpoints (int): Current number of model points.

    Returns:
        dict: Coefficients for updating the ``linear_terms`` and "square_terms"
            of the residual model.
    """
    n = x_sample_monomial_basis.shape[1] - 1
    n_obs = residual_model_interpolated.shape[1]

    params_gradient = np.zeros((n_obs, n))
    params_hessian = np.zeros((n_obs, n, n))
    lower_triangular = lower_triangular[:, n + 1 : n_modelpoints]

    if n_modelpoints == (n + 1):
        omega = np.zeros(n)
        beta = np.zeros(int(n * (n + 1) / 2))
    else:
        lower_triangular_square = np.dot(lower_triangular.T, lower_triangular)

    for k in range(n_obs):
        if n_modelpoints != (n + 1):
            lower_triangular_omega = np.dot(
                basis_null_space[:n_modelpoints, :].T,
                residual_model_interpolated[:n_modelpoints, k],
            )
            omega = np.linalg.solve(
                np.atleast_2d(lower_triangular_square),
                np.atleast_1d(lower_triangular_omega),
            )

            beta = np.dot(np.atleast_2d(lower_triangular), omega)

        rhs = residual_model_interpolated[:n_modelpoints, k] - np.dot(
            monomial_basis[:n_modelpoints, :], beta
        )

        alpha = np.linalg.solve(x_sample_monomial_basis[: n + 1, : n + 1], rhs[: n + 1])
        params_gradient[k, :] = alpha[1 : (n + 1)]

        num = 0
        for i in range(n):
            params_hessian[k, i, i] = beta[num]
            num += 1
            for j in range(i + 1, n):
                params_hessian[k, j, i] = beta[num] / np.sqrt(2)
                params_hessian[k, i, j] = beta[num] / np.sqrt(2)
                num += 1

    coefficients_to_add = {
        "linear_terms": params_gradient.T,
        "square_terms": params_hessian,
    }

    return coefficients_to_add


def update_trustregion_radius(
    result_subproblem,
    rho,
    model_is_valid,
    delta,
    delta_min,
    delta_max,
    eta1,
    gamma0,
    gamma1,
):
    """Update the trust-region radius."""
    norm_x_sub = np.sqrt(np.sum(result_subproblem["x"] ** 2))

    if rho >= eta1 and norm_x_sub > 0.5 * delta:
        delta = min(delta * gamma1, delta_max)
    elif model_is_valid is True:
        delta = max(delta * gamma0, delta_min)

    return delta


def get_last_model_indices_and_check_for_repeated_model(
    model_indices, last_model_indices, n_modelpoints, n_last_modelpoints
):
    """Get the last model_indices and check if we have reused the same model."""
    if n_modelpoints == n_last_modelpoints:
        same_model_used = True
    else:
        same_model_used = False

    for i in range(n_modelpoints):
        if same_model_used:
            if model_indices[i] == last_model_indices[i]:
                same_model_used = True
            else:
                same_model_used = False
        last_model_indices[i] = model_indices[i]

    n_last_modelpoints = n_modelpoints

    return last_model_indices, n_last_modelpoints, same_model_used


def update_model_indices_residual_model(model_indices, accepted_index, n_modelpoints):
    """Update model indices and number of points in the residual model."""
    model_indices[1 : n_modelpoints + 1] = model_indices[:n_modelpoints]
    n_modelpoints += 1
    model_indices[0] = accepted_index

    return model_indices, n_modelpoints


def _get_monomial_basis(x):
    """Get the monomial basis (basis for quadratic functions) of x.

    Monomial basis = .5*[x(1)^2  sqrt(2)*x(1)*x(2) ... sqrt(2)*x(1)*x(n) ...
        ... x(2)^2 sqrt(2)*x(2)*x(3) .. x(n)^2]

    Args:
        x (np.ndarray): Parameter vector of shape (n,).

    Returns:
        np.ndarray: Monomial basis of x wof shape (n * (n + 1) / 2,).
    """
    n = len(x)
    monomial_basis = np.zeros(int(n * (n + 1) / 2))

    j = 0
    for i in range(n):
        monomial_basis[j] = 0.5 * x[i] ** 2
        j += 1

        for k in range(i + 1, n):
            monomial_basis[j] = x[i] * x[k] / np.sqrt(2)
            j += 1

    return monomial_basis



import numpy as np
import pandas as pd
from estimagic.parameters.tree_registry import get_registry
from pybaum import tree_just_flatten


def create_convergence_histories(
    problems, results, stopping_criterion, x_precision, y_precision
):
    """Create tidy DataFrame with all information needed for the benchmarking plots.

    Args:
        problems (dict): estimagic benchmarking problems dictionary. Keys are the
            problem names. Values contain information on the problem, including the
            solution value.
        results (dict): estimagic benchmarking results dictionary. Keys are
            tuples of the form (problem, algorithm), values are dictionaries of the
            collected information on the benchmark run, including 'criterion_history'
            and 'time_history'.
        stopping_criterion (str): one of "x_and_y", "x_or_y", "x", "y". Determines
            how convergence is determined from the two precisions.
        x_precision (float or None): how close an algorithm must have gotten to the
            true parameter values (as percent of the Euclidean distance between start
            and solution parameters) before the criterion for clipping and convergence
            is fulfilled.
        y_precision (float or None): how close an algorithm must have gotten to the
            true criterion values (as percent of the distance between start
            and solution criterion value) before the criterion for clipping and
            convergence is fulfilled.

    Returns:
        pandas.DataFrame: tidy DataFrame with the following columns:
            - problem
            - algorithm
            - n_evaluations
            - walltime
            - criterion
            - criterion_normalized
            - monotone_criterion
            - monotone_criterion_normalized
            - parameter_distance
            - parameter_distance_normalized
            - monotone_parameter_distance
            - monotone_parameter_distance_normalized
    """
    # get solution values for each problem
    registry = get_registry(extended=True)
    x_opt = {
        name: tree_just_flatten(prob["solution"]["params"], registry=registry)
        for name, prob in problems.items()
    }
    f_opt = pd.Series(
        {name: prob["solution"]["value"] for name, prob in problems.items()}
    )

    # build df from results
    time_sr = _get_history_as_stacked_sr_from_results(results, "time_history")
    time_sr.name = "walltime"
    criterion_sr = _get_history_as_stacked_sr_from_results(results, "criterion_history")
    x_dist_sr = _get_history_of_the_parameter_distance(results, x_opt)
    df = pd.concat([time_sr, criterion_sr, x_dist_sr], axis=1)

    df.index = df.index.rename({"evaluation": "n_evaluations"})
    df = df.sort_index().reset_index()

    first_evaluations = df.query("n_evaluations == 0").groupby("problem")
    f_0 = first_evaluations["criterion"].mean()
    x_0_dist = first_evaluations["parameter_distance"].mean()
    x_opt_dist = {name: 0 for name in problems}

    # normalizations
    df["criterion_normalized"] = _normalize(
        df=df, col="criterion", start_values=f_0, target_values=f_opt
    )
    df["parameter_distance_normalized"] = _normalize(
        df=df,
        col="parameter_distance",
        start_values=x_0_dist,
        target_values=x_opt_dist,
    )
    # create monotone versions of columns
    df["monotone_criterion"] = _make_history_monotone(df, "criterion")
    df["monotone_parameter_distance"] = _make_history_monotone(df, "parameter_distance")
    df["monotone_criterion_normalized"] = _make_history_monotone(
        df, "criterion_normalized"
    )
    df["monotone_parameter_distance_normalized"] = _make_history_monotone(
        df, "parameter_distance_normalized"
    )

    if stopping_criterion is not None:
        df, converged_info = _clip_histories(
            df=df,
            stopping_criterion=stopping_criterion,
            x_precision=x_precision,
            y_precision=y_precision,
        )
    else:
        converged_info = None

    return df, converged_info


def _get_history_as_stacked_sr_from_results(results, key):
    """Get history as stacked Series from results.

    Args:
        results (dict): estimagic benchmarking results dictionary.
        key (str): name of the history for which to build the Series, e.g.
            criterion_history.

    Returns:
        pandas.Series: index levels are 'problem', 'algorithm' and 'evaluation'.
            the name is the key with '_history' stripped off.

    """
    histories = {tup: res[key] for tup, res in results.items()}
    sr = pd.concat(histories)
    sr.index.names = ["problem", "algorithm", "evaluation"]
    sr.name = key.replace("_history", "")
    return sr


def _get_history_of_the_parameter_distance(results, x_opt):
    """Calculate the history of the distances to the optimal parameters.

    Args:
        results (dict): estimagic benchmarking results dictionary. Keys are
            tuples of the form (problem, algorithm), values are dictionaries of the
            collected information on the benchmark run, including 'params_history'.
        x_opt (dict): the keys are the problems, the values are pandas.Series with the
            optimal parameters for the respective problem.

    Returns:
        pandas.Series: index levels are "problem", "algorithm", "evaluation". The name
            is "parameter_distance".

    """
    x_dist_history = {}
    for (prob, algo), res in results.items():
        param_history = res["params_history"]

        x_dist_history[(prob, algo)] = pd.Series(
            np.linalg.norm(param_history - x_opt[prob], axis=1)
        )

    sr = pd.concat(x_dist_history)
    sr.index.names = ["problem", "algorithm", "evaluation"]
    sr.name = "parameter_distance"
    return sr


def _make_history_monotone(df, target_col, direction="minimize"):
    """Create a monotone Series, i.e. the best so far instead of the current evaluation.

    Args:
        df (pandas.Dataframe): must contain ["problem", "algorithm", "n_evaluations"]
            and the target_col as columns.
        target_col (str): column of which to create the monotone version.
        direction (str): "minimize" or "maximize". "minimize" makes the history
            monotonically decreasing, "maximize" means the history will be monotonically
            increasing.

    Returns:
        pd.Series: target column where all values that are not weak improvements are
            replaced with the best value so far. Index is the same as that of df.

    """
    sorted_df = df.sort_values(["problem", "algorithm", "n_evaluations"])
    grouped = sorted_df.groupby(["problem", "algorithm"])[target_col]

    if direction == "minimize":
        out = grouped.apply(np.minimum.accumulate)
    elif direction == "maximize":
        out = grouped.apply(np.maximum.accumulate)
    else:
        raise ValueError("Only maximize and minimize are allowed as directions.")

    return out


def _normalize(df, col, start_values, target_values):
    """Normalize the values in **col** relative to the total possible improvement.

    We normalize the values of **col** by calculating the share of the distance between
    the start and target values that is still missing.

    Note: This is correct whether we have a minimization or a maximization problem
    because in the case of a maximization both the sign of the numerator and denominator
    in the line where normalized is defined would be switched, i.e. that cancels out.
    (In the case of a maximization the total improvement would be target - start values
    and the currently still missing improvement would be target - current values)

    Args:
        df (pandas.DataFrame): contains the columns **col** and "problem".
        col (str): name of the column to normalize
        start_values (pandas.Series): index are problems, values are start values
        target_values (pandas.Series): index are problems, values are target values.

    Returns:
        pandas.Series: index is the same as that of sr. The lower the value the closer
            the current value is to the target value. 0 means the target value has been
            reached. 1 means the current value is as far from the target value as the
            start value.

    """
    # expand start and target values to the length of the full DataFrame
    start_values = df["problem"].map(start_values)
    target_values = df["problem"].map(target_values)

    normalized = (df[col] - target_values) / (start_values - target_values)
    return normalized


def _clip_histories(df, stopping_criterion, x_precision, y_precision):
    """Shorten the DataFrame to just the evaluations until each algorithm converged.

    Args:
        df (pandas.DataFrame): index levels are ['problem', 'algorithm', 'evaluation'].
            Columns must include "monotone_criterion_normalized" if stopping_criterion
            includes y and "monotone_parameter_distance_normalized" if x is in
            the stopping_criterion.
        stopping_criterion (str): one of "x_and_y", "x_or_y", "x", "y".
        x_precision (float): when an algorithm's parameters are closer than this to the
            true solution's parameters, the algorithm is counted as having converged.
        y_precision (float): when an algorithm's criterion value is closer than this to
            the solution value, the algorithm is counted as having converged.

    Returns:
        shortened (pandas.DataFrame): the entered DataFrame with all histories
            shortened to stop once conversion according to the given criteria is
            reached.
        converged_info (pandas.DataFrame): columns are the algorithms, index are the
            problems. The values are boolean and True when the algorithm arrived at
            the solution with the desired precision.

    """
    # drop problems with no known solution
    if "x" in stopping_criterion:
        df = df[df["monotone_parameter_distance_normalized"].notnull()]
    if "y" in stopping_criterion:
        df = df[df["monotone_criterion_normalized"].notnull()]

    # determine convergence in the known problems
    if "x" in stopping_criterion:
        x_converged = df["monotone_parameter_distance_normalized"] < x_precision
    if "y" in stopping_criterion:
        y_converged = df["monotone_criterion_normalized"] < y_precision

    # determine converged function evaluations
    if stopping_criterion == "y":
        converged = y_converged
    elif stopping_criterion == "x":
        converged = x_converged
    elif stopping_criterion == "x_and_y":
        converged = y_converged & x_converged
    elif stopping_criterion == "x_or_y":
        converged = y_converged | x_converged
    else:
        raise NotImplementedError(
            f"You specified {stopping_criterion} as stopping_criterion but only the "
            "following are allowed: 'x_and_y', 'x_or_y', 'x', or 'y'."
        )

    first_converged = _find_first_converged(converged, df)

    # keep first converged and non-converged
    shortened = df[~converged | first_converged]

    # create converged_info
    converged.index = pd.MultiIndex.from_frame(df[["problem", "algorithm"]])
    grouped = converged.groupby(["problem", "algorithm"])
    converged_info = grouped.any().unstack("algorithm")

    return shortened, converged_info


def _find_first_converged(converged, df):
    """Identify the first converged entry for each problem run.

    Args:
        converged (pandas.Series): same index as df, True where an algorithm has gotten
            sufficiently close to the solution.
        df (pandas.DataFrame): contains the "problem", "algorithm" and "n_evaluations"
            columns.

    Returns:
        pandas.Series: same index as converged. Only True for the first converged entry
            for each problem run, i.e. problem and algorithm combination.

    """
    # this function can probably be implemented much quicker and easier by shifting
    # the converged Series to identify the first converged entries

    converged_with_multi_index = converged.copy(deep=True)
    multi_index = pd.MultiIndex.from_frame(
        df[["problem", "algorithm", "n_evaluations"]]
    )
    converged_with_multi_index.index = multi_index

    only_converged = converged_with_multi_index[converged_with_multi_index]
    first_true_indices = only_converged.groupby(["problem", "algorithm"]).idxmin()
    first_trues = pd.Series(
        converged_with_multi_index.index.isin(first_true_indices.values),
        index=converged.index,
    )
    return first_trues



import pandas as pd


def check_inputs(data, cluster_by=None, ci_method="percentile", alpha=0.05):
    """Check validity of inputs.
    Args:
        data (pd.DataFrame): original dataset.
        cluster_by (str): column name of variable to cluster by.
        ci_method (str): method of choice for confidence interval computation.
        alpha (float): significance level of choice.

    """

    ci_method_list = ["percentile", "bca", "bc", "t", "normal", "basic"]

    if not isinstance(data, pd.DataFrame):
        raise ValueError("Input 'data' must be DataFrame.")

    elif (cluster_by is not None) and (cluster_by not in data.columns.tolist()):
        raise ValueError(
            "Input 'cluster_by' must be None or a column name of DataFrame."
        )

    elif ci_method not in ci_method_list:
        raise ValueError(
            "ci_method must be 'percentile', 'bc',"
            " 'bca', 't', 'basic' or 'normal', '{method}'"
            " was supplied".format(method=ci_method)
        )

    elif alpha > 1 or alpha < 0:
        raise ValueError("Input 'alpha' must be in [0,1].")



"""Functions to create, run and visualize optimization benchmarks.

TO-DO:
- Add other benchmark sets:
    - finish medium scale problems from https://arxiv.org/pdf/1710.11005.pdf, Page 34.
    - add scalar problems from https://github.com/AxelThevenot
- Add option for deterministic noise or wiggle.
"""
import numpy as np
import pandas as pd
from estimagic import batch_evaluators
from estimagic.optimization.optimize import minimize
from estimagic.optimization.optimize_result import OptimizeResult
from estimagic.parameters.tree_registry import get_registry
from pybaum import tree_just_flatten


def run_benchmark(
    problems,
    optimize_options,
    *,
    batch_evaluator="joblib",
    n_cores=1,
    error_handling="continue",
    seed=None,
):
    """Run problems with different optimize options.

    Args:
        problems (dict): Nested dictionary with benchmark problems of the structure:
            {"name": {"inputs": {...}, "solution": {...}, "info": {...}}}
            where "inputs" are keyword arguments for ``minimize`` such as the criterion
            function and start parameters. "solution" contains the entries "params" and
            "value" and "info" might  contain information about the test problem.
        optimize_options (list or dict): Either a list of algorithms or a Nested
            dictionary that maps a name for optimizer settings
            (e.g. ``"lbfgsb_strict_criterion"``) to a dictionary of keyword arguments
            for arguments for ``minimize`` (e.g. ``{"algorithm": "scipy_lbfgsb",
            "algo_options": {"convergence.relative_criterion_tolerance": 1e-12}}``).
            Alternatively, the values can just be an algorithm which is then benchmarked
            at default settings.
        batch_evaluator (str or callable): See :ref:`batch_evaluators`.
        n_cores (int): Number of optimizations that is run in parallel. Note that in
            addition to that an optimizer might parallelize.
        error_handling (str): One of "raise", "continue".

    Returns:
        dict: Nested Dictionary with information on the benchmark run. The outer keys
            are tuples where the first entry is the name of the problem and the second
            the name of the optimize options. The values are dicts with the entries:
            "params_history", "criterion_history", "time_history" and "solution".
    """
    np.random.seed(seed)

    if isinstance(batch_evaluator, str):
        batch_evaluator = getattr(
            batch_evaluators, f"{batch_evaluator}_batch_evaluator"
        )
    opt_options = _process_optimize_options(optimize_options)

    kwargs_list, names = _get_kwargs_list_and_names(problems, opt_options)

    raw_results = batch_evaluator(
        func=minimize,
        arguments=kwargs_list,
        n_cores=n_cores,
        error_handling=error_handling,
        unpack_symbol="**",
    )

    results = _get_results(names, raw_results, kwargs_list)

    return results


def _process_optimize_options(raw_options):
    if not isinstance(raw_options, dict):
        dict_options = {}
        for option in raw_options:
            if isinstance(option, str):
                dict_options[option] = option
            else:
                dict_options[option.__name__] = option
    else:
        dict_options = raw_options

    out_options = {}
    for name, option in dict_options.items():
        if not isinstance(option, dict):
            option = {"algorithm": option}

        out_options[name] = option

    return out_options


def _get_kwargs_list_and_names(problems, opt_options):
    kwargs_list = []
    names = []

    for prob_name, problem in problems.items():
        for option_name, options in opt_options.items():
            kwargs = {**options, **problem["inputs"]}
            kwargs_list.append(kwargs)
            names.append((prob_name, option_name))

    return kwargs_list, names


def _get_results(names, raw_results, kwargs_list):
    registry = get_registry(extended=True)
    results = {}

    for name, result, inputs in zip(names, raw_results, kwargs_list):

        if isinstance(result, OptimizeResult):
            history = result.history
            params_history = pd.DataFrame(
                [tree_just_flatten(p, registry=registry) for p in history["params"]]
            )
            criterion_history = pd.Series(history["criterion"])
            time_history = pd.Series(history["runtime"])
        elif isinstance(result, str):
            _criterion = inputs["criterion"]

            params_history = pd.DataFrame(
                tree_just_flatten(inputs["params"], registry=registry)
            ).T
            criterion_history = pd.Series(_criterion(inputs["params"])["value"])

            time_history = pd.Series([np.inf])
        else:
            raise ValueError(
                "'result' object is expected to be of type 'dict' or 'str'."
            )

        results[name] = {
            "params_history": params_history,
            "criterion_history": criterion_history,
            "time_history": time_history,
            "solution": result,
        }

    return results



"""Collection of solvers for a quadratic trust-region subproblem."""
import numpy as np
from estimagic.optimization.subsolvers.bounded_newton_quadratic import (
    apply_bounds_to_x_candidate,
)
from estimagic.optimization.subsolvers.bounded_newton_quadratic import (
    check_for_convergence,
)
from estimagic.optimization.subsolvers.bounded_newton_quadratic import (
    compute_conjugate_gradient_step,
)
from estimagic.optimization.subsolvers.bounded_newton_quadratic import (
    compute_predicted_reduction_from_conjugate_gradient_step,
)
from estimagic.optimization.subsolvers.bounded_newton_quadratic import (
    find_hessian_submatrix_where_bounds_inactive,
)
from estimagic.optimization.subsolvers.bounded_newton_quadratic import (
    get_information_on_active_bounds,
)
from estimagic.optimization.subsolvers.bounded_newton_quadratic import (
    take_preliminary_gradient_descent_step_and_check_for_solution,
)
from estimagic.optimization.subsolvers.bounded_newton_quadratic import (
    update_trustregion_radius_conjugate_gradient,
)
from estimagic.optimization.subsolvers.gqtpar_quadratic import (
    add_lambda_and_factorize_hessian,
)
from estimagic.optimization.subsolvers.gqtpar_quadratic import (
    check_for_interior_convergence_and_update,
)
from estimagic.optimization.subsolvers.gqtpar_quadratic import (
    find_new_candidate_and_update_parameters,
)
from estimagic.optimization.subsolvers.gqtpar_quadratic import (
    get_initial_guess_for_lambdas,
)
from estimagic.optimization.subsolvers.gqtpar_quadratic import HessianInfo
from estimagic.optimization.subsolvers.gqtpar_quadratic import (
    update_lambdas_when_factorization_unsuccessful,
)

# ======================================================================================
# Subsolver BNTR
# ======================================================================================


def minimize_bntr_quadratic(
    model,
    lower_bounds,
    upper_bounds,
    *,
    conjugate_gradient_method,
    maxiter,
    maxiter_gradient_descent,
    gtol_abs,
    gtol_rel,
    gtol_scaled,
    gtol_abs_conjugate_gradient,
    gtol_rel_conjugate_gradient,
):
    """Minimize a bounded trust-region subproblem via Newton Conjugate Gradient method.

    The BNTR (Bounded Newton Trust Rregion) algorithm uses an active-set approach
    to solve the symmetric system of equations:

        hessian @ x = - gradient

    only for the inactive parameters of x that lie within the bounds. The active-set
    estimation employed here is based on Bertsekas (:cite:`Bertsekas1982`).

    In the main loop, BNTR globalizes the Newton step using a trust-region method
    based on the predicted versus actual reduction in the criterion function.
    The trust-region radius is increased only if the accepted step is at the
    trust-region boundary.


    Args:
        model (NamedTuple): NamedTuple containing the parameters of the
            main model, i.e.:
            - ``linear_terms`` (np.ndarray): 1d array of shape (n,)
            - ``square_terms`` (np.ndarray): 2d array of shape (n,n).
        lower_bounds (np.ndarray): 1d array of shape (n,) with lower bounds
            for the parameter vector x.
        upper_bounds (np.ndarray): 1d array of shape (n,) with upper bounds
            for the parameter vector x.
        conjugate_gradient_method (str): Method for computing the conjugate gradient
            step. Available conjugate gradient methods are:
                - "cg"
                - "steihaug_toint"
                - "trsbox" (default)
        maxiter (int): Maximum number of iterations. If reached, terminate.
        maxiter_gradient_descent (int): Maximum number of steepest descent iterations
            to perform when the trust-region subsolver BNTR is used.
        gtol_abs (float): Convergence tolerance for the absolute gradient norm.
        gtol_rel (float): Convergence tolerance for the relative gradient norm.
        gtol_scaled (float): Convergence tolerance for the scaled gradient norm.
        gtol_abs_conjugate_gradient (float): Convergence tolerance for the absolute
            gradient norm in the conjugate gradient step of the trust-region
            subproblem ("BNTR").
        gtol_rel_conjugate_gradient (float): Convergence tolerance for the relative
            gradient norm in the conjugate gradient step of the trust-region
            subproblem ("BNTR").

    Returns:
        (dict): Result dictionary containing the following keys:
            - ``x`` (np.ndarray): Solution vector of the subproblem of shape (n,)
            - ``criterion`` (float): Minimum function value associated with the
                solution.
            - ``n_iterations`` (int): Number of iterations the algorithm ran before
                termination.
            - ``success`` (bool): Boolean indicating whether a solution has been found
                before reaching maxiter.
    """
    options_update_radius = {
        "eta1": 1.0e-4,
        "eta2": 0.25,
        "eta3": 0.50,
        "eta4": 0.90,
        "alpha1": 0.25,
        "alpha2": 0.50,
        "alpha3": 1.00,
        "alpha4": 2.00,
        "alpha5": 4.00,
        "min_radius": 1e-10,
        "max_radius": 1e10,
        "default_radius": 100,
    }

    x_candidate = np.zeros_like(model.linear_terms)

    (
        x_candidate,
        f_candidate,
        gradient_unprojected,
        hessian_bounds_inactive,
        trustregion_radius,
        active_bounds_info,
        converged,
        convergence_reason,
    ) = take_preliminary_gradient_descent_step_and_check_for_solution(
        x_candidate,
        model,
        lower_bounds,
        upper_bounds,
        maxiter_gradient_descent,
        gtol_abs,
        gtol_rel,
        gtol_scaled,
    )

    for niter in range(maxiter + 1):
        if converged is True:
            break

        x_old = x_candidate
        f_old = f_candidate
        accept_step = False

        while not accept_step and not converged:
            gradient_bounds_inactive = gradient_unprojected[active_bounds_info.inactive]
            hessian_bounds_inactive = find_hessian_submatrix_where_bounds_inactive(
                model, active_bounds_info
            )
            (
                conjugate_gradient_step,
                conjugate_gradient_step_inactive_bounds,
                cg_step_norm,
            ) = compute_conjugate_gradient_step(
                x_candidate,
                gradient_bounds_inactive,
                hessian_bounds_inactive,
                lower_bounds,
                upper_bounds,
                active_bounds_info,
                trustregion_radius,
                conjugate_gradient_method=conjugate_gradient_method,
                gtol_abs_conjugate_gradient=gtol_abs_conjugate_gradient,
                gtol_rel_conjugate_gradient=gtol_rel_conjugate_gradient,
                options_update_radius=options_update_radius,
            )

            x_unbounded = x_candidate + conjugate_gradient_step
            x_candidate = apply_bounds_to_x_candidate(
                x_unbounded, lower_bounds, upper_bounds
            )

            predicted_reduction = (
                compute_predicted_reduction_from_conjugate_gradient_step(
                    conjugate_gradient_step,
                    conjugate_gradient_step_inactive_bounds,
                    gradient_unprojected,
                    gradient_bounds_inactive,
                    hessian_bounds_inactive,
                    active_bounds_info,
                )
            )

            f_candidate = evaluate_model_criterion(
                x_candidate, model.linear_terms, model.square_terms
            )
            actual_reduction = f_old - f_candidate

            trustregion_radius_old = trustregion_radius
            (
                trustregion_radius,
                accept_step,
            ) = update_trustregion_radius_conjugate_gradient(
                f_candidate,
                predicted_reduction,
                actual_reduction,
                cg_step_norm,
                trustregion_radius,
                options_update_radius,
            )

            if accept_step:
                gradient_unprojected = evaluate_model_gradient(x_candidate, model)

                active_bounds_info = get_information_on_active_bounds(
                    x_candidate,
                    gradient_unprojected,
                    lower_bounds,
                    upper_bounds,
                )
            else:
                x_candidate = x_old
                f_candidate = f_old

                if trustregion_radius == trustregion_radius_old:
                    converged = True
                    break

            converged, convergence_reason = check_for_convergence(
                x_candidate,
                f_candidate,
                gradient_unprojected,
                model,
                lower_bounds,
                upper_bounds,
                converged,
                convergence_reason,
                niter,
                maxiter=maxiter,
                gtol_abs=gtol_abs,
                gtol_rel=gtol_rel,
                gtol_scaled=gtol_scaled,
            )

    result = {
        "x": x_candidate,
        "criterion": f_candidate,
        "n_iterations": niter,
        "success": converged,
        "message": convergence_reason,
    }

    return result


# ======================================================================================
# Subsolver GQTPAR
# ======================================================================================


def minimize_gqtpar_quadratic(model, *, k_easy=0.1, k_hard=0.2, maxiter=200):
    """Solve the quadratic trust-region subproblem via nearly exact iterative method.

    This subproblem solver is mainly based on Conn et al. (2000) "Trust region methods"
    (:cite:`Conn2000`), pp. 169-200.

    But ideas from Nocedal and Wright (2006) "Numerical optimization"
    (:cite:`Nocedal2006`), pp. 83-91, who implement a similar algorithm,
    were also used.

    The original algorithm was developed by More and Sorensen (1983) (:cite:`More1983`)
    and is known as "GQTPAR".

    The vector x* is a global solution to the quadratic subproblem:

        min_x f + g @ x + 0.5 * x.T @ H @ x,

        if and only if ||x|| <= trustregion_radius
        and if there is a scalar lambda >= 0, such that:

    1) (H + lambda * I(n)) x* = -g
    2) lambda (trustregion_radius - ||x*||) = 0
    3) H + lambda * I is positive definite

    where g denotes the gradient and H the hessian of the quadratic model,
    respectively.

    k_easy and k_hard are stopping criteria for the iterative subproblem solver.
    See pp. 194-197 in :cite:`Conn2000` for a more detailed description.

    Args:
        main_model (NamedTuple): NamedTuple containing the parameters of the
            main model, i.e.:
            - ``linear_terms``, a np.ndarray of shape (n,) and
            - ``square_terms``, a np.ndarray of shape (n,n).
        trustregion_radius (float): Trustregion radius, often referred to as delta.
        k_easy (float): topping criterion for the "easy" case.
        k_hard (float): Stopping criterion for the "hard" case.
        maxiter (int): Maximum number of iterations to perform. If reached,
            terminate.

    Returns:
        (dict): Result dictionary containing the following keys:
            - ``x`` (np.ndarray): Solution vector of the subproblem of shape (n,)
            - ``criterion`` (float): Minimum function value associated with the
                solution.
    """
    hessian_info = HessianInfo()

    # Small floating point number signaling that for vectors smaller
    # than that backward substituition is not reliable.
    # See Golub, G. H., Van Loan, C. F. (2013), "Matrix computations", p.165.
    zero_threshold = (
        model.square_terms.shape[0]
        * np.finfo(float).eps
        * np.linalg.norm(model.square_terms, np.Inf)
    )
    stopping_criteria = {
        "k_easy": k_easy,
        "k_hard": k_hard,
    }

    gradient_norm = np.linalg.norm(model.linear_terms)
    lambdas = get_initial_guess_for_lambdas(model)

    converged = False

    for _niter in range(maxiter):

        if hessian_info.already_factorized is True:
            hessian_info = hessian_info._replace(already_factorized=False)
        else:
            hessian_info, factorization_info = add_lambda_and_factorize_hessian(
                model, hessian_info, lambdas
            )

        if factorization_info == 0 and gradient_norm > zero_threshold:
            (
                x_candidate,
                hessian_info,
                lambdas,
                converged,
            ) = find_new_candidate_and_update_parameters(
                model,
                hessian_info,
                lambdas,
                stopping_criteria,
                converged,
            )

        elif factorization_info == 0 and gradient_norm <= zero_threshold:
            x_candidate, lambdas, converged = check_for_interior_convergence_and_update(
                x_candidate,
                hessian_info,
                lambdas,
                stopping_criteria,
                converged,
            )

        else:
            lambdas = update_lambdas_when_factorization_unsuccessful(
                hessian_info,
                lambdas,
                factorization_info,
            )

        if converged is True:
            break

    f_min = evaluate_model_criterion(
        x_candidate, model.linear_terms, model.square_terms
    )

    result = {
        "x": x_candidate,
        "criterion": f_min,
        "n_iterations": _niter,
        "success": converged,
    }

    return result


def evaluate_model_criterion(
    x,
    gradient,
    hessian,
):
    """Evaluate the criterion function value of the main model.

    Args:
        x (np.ndarray): Candidate vector of shape (n,).
        gradient (np.ndarray): Gradient of shape (n,) for which the main model
            shall be evaluated.
        hessian (np.ndarray): Hessian of shape (n, n) for which the main model
            shall be evaulated.

    Returns:
        float: Criterion value of the main model.
    """
    return gradient.T @ x + 0.5 * x.T @ hessian @ x


def evaluate_model_gradient(x, model):
    """Evaluate the derivative of the main model.

    Args:
        x (np.ndarray): Candidate vector of shape (n,).
        main_model (NamedTuple): NamedTuple containing the parameters of the
            main model, i.e.:
            - ``linear_terms``", a np.ndarray of shape (n,) and
            - ``square_terms``, a np.ndarray of shape (n,n).

    Returns:
        np.ndarray: Derivative of the main model of shape (n,).
    """
    return model.linear_terms + model.square_terms @ x



import pandas as pd
from estimagic.batch_evaluators import joblib_batch_evaluator
from estimagic.inference.bootstrap_ci import compute_ci
from estimagic.inference.bootstrap_helpers import check_inputs
from estimagic.inference.bootstrap_outcomes import get_bootstrap_outcomes


def bootstrap(
    data,
    outcome,
    *,
    outcome_kwargs=None,
    n_draws=1_000,
    cluster_by=None,
    ci_method="percentile",
    alpha=0.05,
    seed=None,
    n_cores=1,
    error_handling="continue",
    batch_evaluator=joblib_batch_evaluator,
):
    """Calculate bootstrap estimates, standard errors and confidence intervals
    for statistic of interest in given original sample.

    Args:
        data (pandas.DataFrame): original dataset.
        outcome (callable): function of the data calculating statistic of interest.
            Needs to return a pandas Series.
        outcome_kwargs (dict): Additional keyword arguments for outcome.
        n_draws (int): number of bootstrap samples to draw.
        cluster_by (str): column name of variable to cluster by or None.
        ci_method (str): method of choice for confidence interval computation.
        alpha (float): significance level of choice.
        seeds (numpy.array): array of seeds for bootstrap samples, default is none.
        n_cores (int): number of jobs for parallelization.
        error_handling (str): One of "continue", "raise". Default "continue" which means
            that bootstrap estimates are only calculated for those samples where no
            errors occur and a warning is produced if any error occurs.
        batch_evaluator (str or Callable): Name of a pre-implemented batch evaluator
            (currently 'joblib' and 'pathos_mp') or Callable with the same interface
            as the estimagic batch_evaluators. See :ref:`batch_evaluators`.

    Returns:
        results (pandas.DataFrame): DataFrame where k'th row contains mean estimate,
        standard error, and confidence interval of k'th parameter.

    """

    check_inputs(data, cluster_by, ci_method, alpha)

    estimates = get_bootstrap_outcomes(
        data=data,
        outcome=outcome,
        outcome_kwargs=outcome_kwargs,
        cluster_by=cluster_by,
        seed=seed,
        n_draws=n_draws,
        n_cores=n_cores,
        error_handling=error_handling,
        batch_evaluator=batch_evaluator,
    )

    out = bootstrap_from_outcomes(
        data, outcome, estimates, ci_method=ci_method, alpha=alpha, n_cores=n_cores
    )

    return out


def bootstrap_from_outcomes(
    data, outcome, bootstrap_outcomes, *, ci_method="percentile", alpha=0.05, n_cores=1
):
    """Set up results table containing mean, standard deviation and confidence interval
    for each estimated parameter.

    Args:
        data (pandas.DataFrame): original dataset.
        outcome (callable): function of the data calculating statistic of interest.
            Needs to return a pandas Series.
        bootstrap_outcomes (pandas.DataFrame): DataFrame of bootstrap_outcomes in the
            bootstrap samples.
        ci_method (str): method of choice for confidence interval computation.
        n_cores (int): number of jobs for parallelization.
        alpha (float): significance level of choice.

    Returns:
        results (pandas.DataFrame): table of results.

    """

    check_inputs(data=data, ci_method=ci_method, alpha=alpha)

    summary = pd.DataFrame(bootstrap_outcomes.mean(axis=0), columns=["mean"])

    summary["std"] = bootstrap_outcomes.std(axis=0)

    cis = compute_ci(data, outcome, bootstrap_outcomes, ci_method, alpha, n_cores)
    summary["lower_ci"] = cis["lower_ci"]
    summary["upper_ci"] = cis["upper_ci"]

    cov = bootstrap_outcomes.cov()

    out = {"summary": summary, "cov": cov, "outcomes": bootstrap_outcomes}

    return out
